#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <strings.h>
#include <stdbool.h>
#include <time.h>
#include <inttypes.h>

#include <tox/tox.h>

#include "tox_bot.h"
#include "misc.h"

#define MAX_COMMAND_LENGTH TOX_MAX_MESSAGE_LENGTH
#define MAX_NUM_ARGS 4

/* Commands */
/*          */
static void cmd_help(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH], bool user)
{
    char com_res[TOX_MAX_MESSAGE_LENGTH];
    char str1[100] = "\n\n\t\t### Command list ### \n";
    char str2[100] = "ping\t\t\t\t\t - Print pong answer\n";
    char str3[100] = "echo \"<message>\"\t\t\t - Echoing your message\n";
    char str4[100] = "changename \"<new bot name>\"\t\t - Set a new bot name\n";
    char str5[100] = "changestatus \"<new bot statusname>\"\t - Set a new bot status message\n";
    snprintf(com_res, sizeof com_res, "%s%s%s%s%s", str1, str2, str3, str4, str5);

    tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) com_res, strlen(com_res), NULL);
    log_msg(tox, com_res, friendnum, user);
}

/* -------------------------------------------------------------------------- */

static void cmd_ping(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH], bool user)
{
    const char *outmsg = NULL;

    outmsg = "pong";

    tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) outmsg, strlen(outmsg), NULL);
    log_msg(tox, outmsg, friendnum, user);
}

/* -------------------------------------------------------------------------- */

static void cmd_echo(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH], bool user)
{
    const char *inmsg = NULL;

    if (argc < 1) {
        inmsg = "Error: Message required";
        tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) inmsg, strlen(inmsg), NULL);
        return;
    }

    char outmsg[TOX_MAX_MESSAGE_LENGTH];
    int len = 0;

    if (argv[1][0] == '\"') {    /* remove opening and closing quotes */
        snprintf(outmsg, sizeof(outmsg), "%s", &argv[1][1]);
        len = strlen(outmsg) - 1;
    } else {
        snprintf(outmsg, sizeof(outmsg), "%s", argv[1]);
        len = strlen(outmsg);
    }

    outmsg[len] = '\0';

    tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) outmsg, strlen(outmsg), NULL);
    log_msg(tox, outmsg, friendnum, user);
}

/* -------------------------------------------------------------------------- */

static void bot_changename(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH], bool user)
{
    const char *outmsg = NULL;

    if (argc < 1) {
        outmsg = "Error: Name required";
        tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) outmsg, strlen(outmsg), NULL);
        return;
    }

    char name[TOX_MAX_NAME_LENGTH];
    int len = 0;

    if (argv[1][0] == '\"') {    /* remove opening and closing quotes */
        snprintf(name, sizeof(name), "%s", &argv[1][1]);
        len = strlen(name) - 1;
    } else {
        snprintf(name, sizeof(name), "%s", argv[1]);
        len = strlen(name);
    }

    name[len] = '\0';
    tox_self_set_name(tox, (uint8_t *) name, (uint16_t) len, NULL);

    char com_res[TOX_MAX_MESSAGE_LENGTH];
    snprintf(com_res, sizeof com_res, "%s%s%s", "*** Name changed to \"", name, "\" ***");

    tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) com_res,
                            strlen(com_res), NULL);
    write_conf(tox, name, NULL);
    log_msg(tox, com_res, friendnum, user);

}

/* -------------------------------------------------------------------------- */

static void bot_changestatus(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH], bool user)
{
    const char *outmsg = NULL;

    if (argc < 1) {
        outmsg = "Error: Status required";
        tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) outmsg, strlen(outmsg), NULL);
        return;
    }

    char status[TOX_MAX_STATUS_MESSAGE_LENGTH];
    int len = 0;

    if (argv[1][0] == '\"') {    /* remove opening and closing quotes */
        snprintf(status, sizeof(status), "%s", &argv[1][1]);
        len = strlen(status) - 1;
    } else {
        snprintf(status, sizeof(status), "%s", argv[1]);
        len = strlen(status);
    }

    status[len] = '\0';

    tox_self_set_status_message(tox, (uint8_t *) status, (uint16_t) len, NULL);

    char com_res[TOX_MAX_MESSAGE_LENGTH];
    snprintf(com_res, sizeof com_res, "%s%s%s", "*** Status changed to \"", status, "\" ***");

    tox_friend_send_message(tox, friendnum, TOX_MESSAGE_TYPE_NORMAL, (uint8_t *) com_res,
                            strlen(com_res), NULL);
    write_conf(tox, NULL, status);
    log_msg(tox, com_res, friendnum, user);

}

/*              */
/* Commands end */


/* Parses input command and puts args into arg array.
   Returns number of arguments on success, -1 on failure. */
static int parse_command(const char *input, char (*args)[MAX_COMMAND_LENGTH])
{
    char *cmd = strdup(input);

    if (cmd == NULL) {
        exit(EXIT_FAILURE);
    }

    int num_args = 0;
    int i = 0;    /* index of last char in an argument */

    /* characters wrapped in double quotes count as one arg */
    while (num_args < MAX_NUM_ARGS) {
        int qt_ofst = 0;    /* set to 1 to offset index for quote char at end of arg */

        if (*cmd == '\"') {
            qt_ofst = 1;
            i = char_find(1, cmd, '\"');

            if (cmd[i] == '\0') {
                free(cmd);
                return -1;
            }
        } else {
            i = char_find(0, cmd, ' ');
        }

        memcpy(args[num_args], cmd, i + qt_ofst);
        args[num_args++][i + qt_ofst] = '\0';

        if (cmd[i] == '\0') {  /* no more args */
            break;
        }

        char tmp[MAX_COMMAND_LENGTH];
        snprintf(tmp, sizeof(tmp), "%s", &cmd[i + 1]);
        strcpy(cmd, tmp);    /* tmp will always fit inside cmd */
    }

    free(cmd);
    return num_args;
}

/* -------------------------------------------------------------------------- */

static struct {
    const char *name;
    void (*func)(Tox *tox, uint32_t friendnum, int argc, char (*argv)[MAX_COMMAND_LENGTH],
                bool user);
} commands[] = {
    { "help",            cmd_help          },
    { "ping",            cmd_ping          },
    { "echo",            cmd_echo          },
    { "changename",      bot_changename    },
    { "changestatus",    bot_changestatus  },
    { NULL,               NULL             },
};

/* -------------------------------------------------------------------------- */

static int do_command(Tox *tox, uint32_t friendnum, int num_args, char (*args)[MAX_COMMAND_LENGTH],
                      bool user)
{
    int i;

    for (i = 0; commands[i].name; ++i) {
        if (strcmp(args[0], commands[i].name) == 0) {
            (commands[i].func)(tox, friendnum, num_args - 1, args, user);
            return 0;
        }
    }

    return -1;
}

/* -------------------------------------------------------------------------- */

int execute(Tox *tox, uint32_t friendnum, const char *input, int length, bool user)
{
    if (length >= MAX_COMMAND_LENGTH) {
        return -1;
    }

    char args[MAX_NUM_ARGS][MAX_COMMAND_LENGTH];
    int num_args = parse_command(input, args);

    if (num_args == -1) {
        return -1;
    }

    return do_command(tox, friendnum, num_args, args, user);
}
